import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.*;
 /**
 * CREDITS:
 *    Enemy death sound effect comes from https://www.youtube.com/watch?v=cVonclIw-u0
 *    My brother wrote the following code segments:
 *      - helper to mark the path in Grid given a start and end point.
 *      - bug fix to prevent refrence errors with the path.
 *      - tower targeting method.
 *
 * INSTRUCTIONS:
 *    - Click to select a tower at the bottom of the game screen, and then click
 *    somehwere on the grid to build a tower at that location.
 *
 *    - Press the spacebar to start a wave of enemies.
 *
 *    - Click on existing towers to access the upgrade menu, upgrade will upgrade
 *    the damage and the range, sell will sell the tower and return some money
 *    back to the player, and cancel exits the menu.
 *
 *    - The goal is to reach as high of a level as possible.
 *    - If the player lives reaches 0 the game will end.
 *    - Enemies will get stronger as the levels increase.
 *    - Every 20 levels a boss will spawn with more significantly health.
 *
 *
 * KNOWN BUGS:
 * Screen may flicker in the main menu/start screen.
 *
 */

public class MyWorld extends World
{
  private Grid grid;
  private Spawner spawner;


  private ArrayList<Coord> path;
  private Set<Integer> enemyIdsOnMap;

  private GreenfootImage background;

  //variables for coordinating timing
  private boolean spawnPhase;
  private int frame;

  //game instance variables
  private int level;
  private int playerLives;
  private int score;
  private int money;


  //Labels for UI (not clickeable)
  private UIText levelUI;
  private UIText playerHealthUI;
  private UIText currencyUI;
  private UIText scoreUI;
  private UIText shotLevelLabel;
  private UIText rangeLabel;
  private UIText upgradeCostLabel;

 //action Buttons
  private UIButton upgradeButton;
  private UIButton sellButton;
  private UIButton cancelButton;

// building tower buttons
  private UITowerButton basicTowerButton;
  private UITowerButton speedTowerButton;
  private UITowerButton sniperTowerButton;


// constant to define game dimentions
  private static final int GAME_WIDTH = 960;
  private static final int GAME_HEIGHT = 540;
  private static final int GAME_GRID_END = 400;

  //header Y for the UI positioning
  private static final int HEADER_Y = GAME_GRID_END;

  private boolean gameStart;


  private TowerInfo selectedTower;
  private Tower selectedTowerOnGrid;

    public MyWorld()
    {
        super(GAME_WIDTH, GAME_HEIGHT, 1);
        //create a new Grid object filling the world with 10x20 cells
        //MAP3 is a default map stored in Grid
        this.grid = new Grid(GAME_WIDTH, GAME_GRID_END, 10, 20, Grid.MAP3);

        //generate the image for the grid
        GreenfootImage gridImage = grid.getGridImage();
        //retrieve the path from the grid
        this.path = grid.getPath();

        //construct the background for the world
        //background is made up of the grid image + a UI section
        background = new GreenfootImage(GAME_WIDTH, GAME_HEIGHT);
        background.drawImage(gridImage, 0, 0);
        background.setColor(Color.WHITE);
        background.fillRect(0, GAME_GRID_END, GAME_WIDTH, (GAME_HEIGHT - GAME_GRID_END));
        setBackground(background);

        //set instance variables to default values
        this.spawnPhase = false;
        enemyIdsOnMap = new HashSet<Integer>();
        this.level = 1;
        this.score = 0;
        this.money = 100;
        this.playerLives = 25;
        //construct a spawner for the first level
        this.spawner = new Spawner(level);

        //assign the default first tower to build to be BasicTowers
        selectedTower = BasicTower.getTowerInfo();

        //since there is no towers on map, default is null
        this.selectedTowerOnGrid = null;
        //the game has not started yet
        this.gameStart = false ;
        //load the UI
        loadUI();
        //load the Start screen and cover the baic UI
        UIFullScreen startScreen = new UIFullScreen(GAME_WIDTH, GAME_HEIGHT, "Grid Defense", "Click to start", true );
        addObject(startScreen, GAME_WIDTH/2, GAME_HEIGHT/2);

    }


//________________________UI METHODS_________________________________

    //helper to load the UI
    private void loadUI(){
      //clear any old UI elements first
      clearUI();
      //load in the stats UI
      loadStats();
      //if a tower is not selected load in the tower build UI
      if(selectedTowerOnGrid == null){
        basicTowerButton = new UITowerButton(100, 70, BasicTower.getTowerInfo());
        speedTowerButton = new UITowerButton(100, 70, SpeedTower.getTowerInfo());
        sniperTowerButton = new UITowerButton(100, 70, SniperTower.getTowerInfo());

        // y = GRID_END + the height of the Header bar + margin + half the height of the current elements
        int yLocation = HEADER_Y + levelUI.getHeight() + 10;
        addObject(basicTowerButton, 225, yLocation + (basicTowerButton.getHeight() / 2));
        addObject(speedTowerButton, 400, yLocation +  (speedTowerButton.getHeight() / 2));
        addObject(sniperTowerButton, 575, yLocation + (sniperTowerButton.getHeight()/2));
      }
      //if the tower is selected, load in that stats and upgrade UI for tower
      else {
        //create all the Buttons
        upgradeButton = new UIButton(UIButton.UPGRADE, Color.CYAN);
        sellButton = new UIButton(UIButton.SELL, Color.GREEN);
        cancelButton = new UIButton(UIButton.CANCEL, Color.RED);
        //create all the Labels
        shotLevelLabel = new UIText(selectedTowerOnGrid.getShotLevel(), "Shot Level: ");
        rangeLabel = new UIText(selectedTowerOnGrid.getRange(),"Range: ");
        upgradeCostLabel = new UIText(selectedTowerOnGrid.getUpgradePrice(), "Upgrade Cost: ");
        //Y position is Bottom of grid + any height taken by stats UI + padding


        final int ACTION_BUTTON_Y = HEADER_Y + levelUI.getHeight() + 30;
        int yLocation = ACTION_BUTTON_Y;
        addObject(upgradeButton,225, ACTION_BUTTON_Y + (upgradeButton.getHeight() / 2) );
        addObject(sellButton, 400,  ACTION_BUTTON_Y + (sellButton.getHeight() / 2) );
        addObject(cancelButton, 575,  ACTION_BUTTON_Y + (cancelButton.getHeight() / 2) );
        //iterate the yLocation down so UI can be stacked
        addObject(shotLevelLabel, 700, yLocation + (shotLevelLabel.getHeight()/2));
        yLocation += (shotLevelLabel.getHeight() + 10);
        addObject(rangeLabel, 700, yLocation + (rangeLabel.getHeight()/2));
        yLocation += (rangeLabel.getHeight()+ 10);
        addObject(upgradeCostLabel, 700, yLocation + (upgradeCostLabel.getHeight()/2));

      }
    }

    //helper to load the Stats UI (score , lives, currency, etc.)
    private void loadStats(){
      //constant to line up all status bar objects
      levelUI = new UIText(this.level, "Level: ");
      playerHealthUI = new UIText(this.playerLives, "Lives: ");
      currencyUI =  new UIText(this.money, "Money: ");
      scoreUI = new UIText(this.score, "Score: ");
      addObject(levelUI, 225 ,HEADER_Y + levelUI.getHeight()/2);
      addObject(playerHealthUI, 425, HEADER_Y + playerHealthUI.getHeight()/2);
      addObject(currencyUI,625, HEADER_Y + currencyUI.getHeight()/2);
      addObject(scoreUI, 825, HEADER_Y + scoreUI.getHeight()/2);
    }

    //function to clear all UI elements
    private void clearUI(){
      //clear all the Tower build Buttons
      for (UITowerButton buildButtons : getObjects(UITowerButton.class)){
        removeObject(buildButtons);
      }
      //clear all Buttons
      for (UIButton actionButton : getObjects(UIButton.class)){
        removeObject(actionButton);
      }
      //clear all labels
      for (UIText labels : getObjects(UIText.class)){
        removeObject(labels);
      }
    }


//_________________ACT METHOD AND HELPERS__________________________________




  public void act(){
    //ends the act method so that the game can't function
    if (!gameStart) return;
    //run the timing method to spawn enemies
    spawnEnemyByInterval(frame, 30);
    //check the input
    checkInput();
    frame++;
  }


  //helper to check input from the user
  private void checkInput(){
    //if the mouse is clicked
    if (Greenfoot.mouseClicked(null)){
      MouseInfo m = Greenfoot.getMouseInfo();
      if (m != null){
        //convert mouse x,y to Coord object
        Coord pos = new Coord (m.getX(), m.getY());
        //if that tower spot is buildable, construct tower
        if (!grid.isOutOfBounds(pos)){
          buildTower(pos);
        }
      }
    }
    //if the space button is pressed start the wave
    if (Greenfoot.isKeyDown("space")){
      this.spawnPhase = true;
    }
  }


  //helper to time the enemy spawning
  private void spawnEnemyByInterval(int curFrame, int interval){
      //if you are in spawn mode and the frame interval and you have not spawned all your enemies
      if (spawnPhase && !spawner.isEmpty() && curFrame % interval == 0){
        //spawner to spawn the next enemy in queue
          Enemy e = spawner.getEnemy(this.path);
          //add the id to the set
          enemyIdsOnMap.add(e.getID());
          //add to the game at the first coord of the path
          addObject(e, this.path.get(0).getX(), this.path.get(0).getY());
      }
  }


//______________METHODS CALLED BY OTHER OBJECTS___________________________



    //what to do when enemy has reached the end zone
    //called by Enemy Objects
    public void enemyGoalReached(Enemy e){
      //remove that enemy from the set since it it no longer on grid
      enemyIdsOnMap.remove(e.getID());
      //remove object
      removeObject(e);
      //check if the round is over
      checkRoundStatus();
      //remove from the player's lives
      playerLives -= e.getPlayerLivesLost();
      //update the UI
      playerHealthUI.updateValue(playerLives);
      //if player lost all their lives end the game
      if (playerLives <= 0){
        endGame();
      }
    }

    //waht to do  when Enemey is killed by a tower
    public void enemyIsKilled(Enemy e){
      //add the score and currency
      this.score += e.getScore();
      this.money += e.getCoinValue();
      //remove the ID of that enemy and check the round status
      enemyIdsOnMap.remove(e.getID());
      removeObject(e);
      checkRoundStatus();
      //update the UI
      scoreUI.updateValue(this.score);
      currencyUI.updateValue(this.money);
    }


    //method to determine what action to do, this is called by UIButton Objects
    public void buttonPressed(String action){
      switch(action){
        case UIButton.SELL:
          sellSelectedTower();
          break;
        case UIButton.UPGRADE:
          upgradeSelectedTower();
          loadUI();
          break;
        case UIButton.CANCEL:
          selectedTowerOnGrid = null;
          loadUI();
          break;
      }
    }

    //method to assign that a tower on the grid is selected
    //called by Tower Objects
    public void setSelectedTowerOnGrid(Tower t){
      this.selectedTowerOnGrid = t;
      loadUI();
    }

    //method to assign the current Tower to built
    //called by UITowerButton objects
    public void setSelectedTower(TowerInfo t){
      this.selectedTower = t;
    }

    //method to start the game
    //called by Start Screen object
    public void startGame(){
      this.gameStart = true;
    }


//________________________HELPER METHODS_____________________________________



    //check if the round is over
    private void checkRoundStatus(){
      //if all enemies spawned in have despawned
      //and there are no more enemies left to spawn
      if(enemyIdsOnMap.isEmpty() && spawner.isEmpty()){
        //stop the current mode
        spawnPhase = false;
        //increase level
        this.level++;
        //update the UI
        levelUI.updateValue(this.level);
        //build the Spawner for th enext level
        spawner = new Spawner(level);
      }
    }

    //helper to sell the selected Tower
    private void sellSelectedTower(){
      //add back 60% of the tower's value
      this.money += (selectedTowerOnGrid.getPrice()*0.6);
      //update the UI
      currencyUI.updateValue(this.money);
      Tower t = selectedTowerOnGrid;
      //mark the spot as buildable again
      grid.setOccupied(t.getCoord(),false);
      //set tower selected to null
      selectedTowerOnGrid = null;
      //reload the UI
      loadUI();
      //remove tower on grid
      removeObject(t);

    }
    //helper to upgrade tower
    private void upgradeSelectedTower(){
      //get the price to upgrade
      int price = selectedTowerOnGrid.getUpgradePrice();
      if (price <= this.money){
        this.money -= price;
        //call the tower's upgrade method
        selectedTowerOnGrid.upgrade();
        //update the UI with the upgraded stats
        shotLevelLabel.updateValue(selectedTowerOnGrid.getShotLevel());
        rangeLabel.updateValue(selectedTowerOnGrid.getRange());
      }
    }



    //helper to construct the tower
    private void buildTower(Coord pos){
      //get the price of the tower button currently selected
      int price = selectedTower.getPrice();
      //if you clicked in the grid and that spot is buildable
      if ((!grid.isOutOfBounds(pos)) && grid.isBuildable(pos)){

        //get the center of the tile you clicked on
        Coord snapPoint = grid.getSnapPoint(pos);
        if (this.money >= price){
          this.money -= price;
          //updtate the UI with the new value
          currencyUI.updateValue(this.money);
          //TowerSpawner creates the proper subclass based on towerInfo
          Tower newTower = TowerSpawner.spawn(selectedTower, snapPoint.getX(), snapPoint.getY(), 1);
          //add to the wrold
          addObject(newTower, snapPoint.getX(), snapPoint.getY());
          // tell the grid that spot is occupied
          grid.setOccupied(snapPoint ,true);
        }
      }
    }

    //helper to end the game
    public void endGame(){
      //stop the game
      this.gameStart = false;
      //build the End screen
      UIFullScreen endGameScreen = new UIFullScreen(GAME_WIDTH, GAME_HEIGHT, "Game Over", "Score: " + Integer.toString(this.score), false );
      addObject(endGameScreen, GAME_WIDTH/2, GAME_HEIGHT/2);
    }


//_________________________GETTERS________________________________________

      //give access to grid Object
      public Grid getGrid(){
        return grid;
      }


}
